// https://excalidraw.com/#json=mgBWynet5WUbpgfSd0pzC,TMhzt6L-x5yShNCHRGKLvg
extern "C" __global__ void common_mode_asic(
    float* const data,
    const unsigned short n_frames,
    const unsigned short num_iter,
    const float min_dark_fraction,
    const float noise_peak_range
) {
    const int frame = blockIdx.z;
    const int ss_asic = blockIdx.y;
    const int fs_asic = blockIdx.x;
    const int thread_in_asic = threadIdx.x;
    assert(blockDim.x == {{asic_dim}});

    __shared__ float shared_sum[{{asic_dim}}];
    __shared__ unsigned int shared_count[{{asic_dim}}];

    const unsigned int min_dark_pixels = static_cast<int>(min_dark_fraction * static_cast<float>({{asic_dim}} * {{asic_dim}}));
    float* const asic_start = data + frame * {{ss_dim}} * {{fs_dim}} + ss_asic * {{asic_dim}} * {{fs_dim}} + fs_asic * {{asic_dim}}; 

    for (int iter=0; iter<num_iter; ++iter) {
        // accumulate (over rows) to shared
        unsigned int my_count = 0;
        float my_sum = 0;
        for (int row=0; row<{{asic_dim}}; ++row) {
            const float pixel = asic_start[row * {{fs_dim}} + thread_in_asic];
            if (!isfinite(pixel) || fabs(pixel) > noise_peak_range) {
                continue;
            }
            my_sum += pixel;
            ++my_count;
        }
        shared_sum[thread_in_asic] = my_sum;
        shared_count[thread_in_asic] = my_count;
        __syncthreads();

        // reduce (per ASIC) in shared
        int active_threads = {{asic_dim}};
        while (active_threads > 1) {
            active_threads /= 2;
            const int other = thread_in_asic + active_threads;
            if (thread_in_asic < active_threads) {
                shared_sum[thread_in_asic] += shared_sum[other];
                shared_count[thread_in_asic] += shared_count[other];
            }
            __syncthreads();
        }

        // now, each ASIC result should be in first index of shared by ASIC
        const unsigned int dark_count = shared_count[0];
        if (dark_count < min_dark_pixels) {
            return;
        }
        const float dark_sum = shared_sum[0];
        const float dark_mean = dark_sum / static_cast<float>(dark_count);
        for (int row=0; row<{{asic_dim}}; ++row) {
	        asic_start[row * {{fs_dim}} + thread_in_asic] -= dark_mean;
        }
    }
}
